from .dt_command_abs import DTCommandAbs

__all__ = ["DTCommandPlaceholder"]


class DTCommandPlaceholder(DTCommandAbs):
    fake = True

    @staticmethod
    def command(shell, args):
        return



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod

__all__ = ["DTCommandAbs"]


class DTCommandAbs(metaclass=ABCMeta):
    name = None
    level = None
    help = None
    commands = None
    fake = False

    @staticmethod
    @abstractmethod
    def command(shell, args):
        pass

    @staticmethod
    def complete(shell, word, line):
        return []

    @staticmethod
    def fail(msg):
        raise Exception(msg)

    @staticmethod
    def do_command(cls, shell, line):
        # print('>[%s]@(%s, %s)' % (line, cls.name, cls.__class__))
        line = line.strip()
        parts = [p.strip() for p in line.split(" ")]
        args = [p for p in parts if len(p) > 0]
        from .utils import undo_replace_spaces

        args = list(map(undo_replace_spaces, args))
        word = parts[0]
        # print('[%s, %r]@(%s, %s)' % (word, parts, cls.name, cls.__class__))
        if len(word) > 0:
            if len(cls.commands) > 0:
                if word in cls.commands:
                    cls.commands[word].do_command(cls.commands[word], shell, " ".join(parts[1:]))
                else:
                    print(
                        "Command `%s` not recognized.\nAvailable sub-commands are:\n\n\t%s"
                        % (word.strip(), "\n\t".join(cls.commands.keys()))
                    )
            else:
                cls.command(shell, args)
        else:
            if len(cls.commands) > 0:
                print("Available sub-commands are:\n\n\t%s" % "\n\t".join(cls.commands.keys()))
            else:
                if not cls.fake:
                    cls.command(shell, args)

    @staticmethod
    def complete_command(cls, shell, word, line, start_index, end_index):
        # print('[%s](%s)@(%s, %s)' % (word, line, cls.name, cls.__class__))
        word = word.strip()
        line = line.strip()
        subcmds = cls.commands.keys()
        parts = [p.strip() for p in line.split(" ")]
        #
        partial_word = len(word) != 0
        if parts[0] == cls.name:
            if len(parts) == 1 or (len(parts) == 2 and partial_word):
                static_comp = [
                    k for k in cls.complete(shell, word, line) if (not partial_word or k.startswith(word))
                ]
                comp_subcmds = static_comp + [k for k in subcmds if (not partial_word or k.startswith(word))]
                # print '!T'
                return comp_subcmds
            if len(parts) > 1 and parts[1] in cls.commands.keys():
                child = parts[1]
                nline = " ".join(parts[1:])
                # print '!C'
                return DTCommandAbs.complete_command(
                    cls.commands[child], shell, word, nline, start_index, end_index
                )
        # print '!D'
        return []

    @staticmethod
    def help_command(cls, shell):
        msg = cls.help if (cls.level == 0 and cls.help is not None) else str(shell.nohelp % cls.name)
        print(msg)



from .exceptions import UserError

__all__ = ['check_package_version']

def check_package_version(PKG: str, min_version: str):
    pip_version = "?"
    try:
        # noinspection PyCompatibility
        from pip import __version__

        pip_version = __version__
        # noinspection PyCompatibility
        from pip._internal.utils.misc import get_installed_distributions
    except ImportError:
        msg = f"""
           You need a higher version of "pip".  You have {pip_version}

           You can install it with a command like:

               pip install -U pip

           (Note: your configuration might require a different command.)
           """
        raise UserError(msg)

    installed = get_installed_distributions()
    pkgs = {_.project_name: _ for _ in installed}
    if PKG not in pkgs:
        msg = f"""
        You need to have an extra package installed called `{PKG}`.

        You can install it with a command like:

            pip3 install -U "{PKG}>={min_version}"

        (Note: your configuration might require a different command.
         You might need to use "pip" instead of "pip3".)
        """
        raise UserError(msg)

    p = pkgs[PKG]

    installed_version = parse_version(p.version)
    required_version = parse_version(min_version)
    if installed_version < required_version:
        msg = f"""
       You need to have installed {PKG} of at least {min_version}.
       We have detected you have {p.version}.

       Please update {PKG} using pip.

           pip3 install -U  "{PKG}>={min_version}"

       (Note: your configuration might require a different command.
        You might need to use "pip" instead of "pip3".)
       """
        raise UserError(msg)


def parse_version(x):
    return tuple(int(_) for _ in x.split("."))



import grp
import os
import subprocess
import sys
from typing import List, Optional, Tuple

from whichcraft import which

from .config import read_shell_config
from .exceptions import InvalidEnvironment, UserError


def running_with_sudo() -> bool:
    if "SUDO_USER" in os.environ:
        return True
    return False


def abort_if_running_with_sudo() -> None:
    if running_with_sudo():
        msg = """\
Do not run dts using "sudo".'

As a matter of fact, do not run anything with "sudo" unless instructed to do so.\
"""
        raise UserError(msg)


def check_docker_environment():
    """ Returns docker client """

    from . import dtslogger

    # dtslogger.debug('Checking docker environment for user %s' % username)

    check_executable_exists("docker")

    check_user_in_docker_group()
    #
    # if on_linux():
    #
    #     if username != 'root':
    #         check_user_in_docker_group()
    #     # print('checked groups')
    # else:
    #     dtslogger.debug('skipping env check because not on Linux')

    try:
        import docker
    except Exception as e:
        msg = "Could not import package docker:\n%s" % e
        msg += "\n\nYou need to install the package"
        raise InvalidEnvironment(msg)

    if "DOCKER_HOST" in os.environ:
        msg = 'Note that the variable DOCKER_HOST is set to "%s"' % os.environ["DOCKER_HOST"]
        dtslogger.warning(msg)

    try:
        # noinspection PyUnresolvedReferences
        client = docker.from_env()

        # TODO: why are we doing this? It seems expensive and useless
        # _containers = client.containers.list(filters=dict(status="running"))

        # dtslogger.debug(json.dumps(client.info(), indent=4))

    except Exception as e:
        msg = "I cannot communicate with Docker:\n%s" % e
        msg += "\n\nMake sure the docker service is running."
        raise InvalidEnvironment(msg)

    return client


def on_linux() -> bool:
    return sys.platform.startswith("linux")


def check_executable_exists(cmdname: str) -> None:
    p = which(cmdname)
    if p is None:
        msg = 'Could not find executable "%s".' % cmdname
        raise InvalidEnvironment(msg)


def check_user_in_docker_group() -> None:
    # first, let's see if there exists a group "docker"
    group_names = [g.gr_name for g in grp.getgrall()]
    G = "docker"
    if G not in group_names:
        pass
    else:
        group_id = grp.getgrnam(G).gr_gid
        my_groups = os.getgroups()
        if group_id not in my_groups:
            msg = 'My groups are %s and "%s" group is %s ' % (my_groups, G, group_id)
            msg += "\n\nNote that when you add a user to a group, you need to login in and out."
            # dtslogger.debug(msg)


def get_active_groups(username: Optional[str] = None) -> List[str]:
    cmd = ["groups"]

    if username:
        cmd.append(username)

    try:
        stdout = subprocess.check_output(["groups"])
    except subprocess.CalledProcessError as e:
        raise Exception(str(e))
    active_groups = stdout.decode().split()

    return active_groups


def get_dockerhub_username() -> str:
    """ raise InvalidEnvironment """

    msg = """
        Please set docker username using
         
            dts challenges config --docker-username <USERNAME>
            
    """
    try:
        shell_config = read_shell_config()
    except Exception:
        raise

    if shell_config.docker_username is None:
        raise InvalidEnvironment(msg)

    return shell_config.docker_username


def get_dockerhub_username_and_password() -> Tuple[str, str]:
    """ raise InvalidEnvironment """
    try:
        shell_config = read_shell_config()
    except Exception:
        raise

    if shell_config.docker_username is None:
        msg = """
    Please set DockerHub username and password/token using:
    
       dts challenges config --docker-username <USERNAME> --docker-password <PASSWORD or DOCKERHUB TOKEN>
       
    You can use your DockerHub password or use a token that you can obtain at 
    
        https://hub.docker.com/settings/security              
        
    (note: this is not the Duckietown token)
    """

        raise InvalidEnvironment(msg)

    elif shell_config.docker_password is None:
        msg = """
   Please set DockerHub username and password/token using:

       dts challenges config --docker-password <PASSWORD or DOCKERHUB TOKEN>
    
    You can use your DockerHub password or use a DockerHub token that you can obtain at 
        
            https://hub.docker.com/settings/security 
            
    (note: this is not the Duckietown token)
       """

        raise InvalidEnvironment(msg)

    return shell_config.docker_username, shell_config.docker_password



# -*- coding: utf-8 -*-
import logging

logging.basicConfig()

dtslogger = logging.getLogger("dts")
dtslogger.setLevel(logging.INFO)

__version__ = "5.2.4"

dtslogger.debug(f"duckietown-shell {__version__}")

import sys

if sys.version_info < (3, 6):
    msg = f"! duckietown-shell works with Python 3.6 and later !.\nDetected {sys.version}."
    logging.error(msg)
    sys.exit(2)

from .exceptions import ConfigInvalid, ConfigNotPresent
from .utils import format_exception

import locale

dtslogger.debug(
    f"encoding: stdout {sys.stdout.encoding} stderr {sys.stderr.encoding} "
    f"locale {locale.getpreferredencoding()}."
)

from .cli import DTShell
from .logging import dts_print

from .dt_command_abs import DTCommandAbs
from .dt_command_placeholder import DTCommandPlaceholder
from .main import cli_main
from .exceptions import *

from .main import OtherVersions
from .package_version_check import *



import argparse
from dataclasses import dataclass
from typing import List, Optional, Tuple

from .constants import ALLOWED_BRANCHES

__all__ = ["CLIOptions", "get_cli_options"]


@dataclass
class CLIOptions:
    debug: bool
    set_version: Optional[str]
    quiet: bool


def get_cli_options(args: List[str]) -> Tuple[CLIOptions, List[str]]:
    """ Returns cli options plus other arguments for the commands. """
    allowed_branches = [b.split("(")[0] for b in ALLOWED_BRANCHES]

    if args and not args[0].startswith("-"):
        return CLIOptions(debug=False, set_version=None, quiet=False), args
    parser = argparse.ArgumentParser()

    parser.add_argument("--debug", action="store_true", default=False, help="More debug information")
    parser.add_argument("-q", "--quiet", action="store_true", default=False, help="Quiet execution")
    parser.add_argument(
        "--set-version",
        type=str,
        default=None,
        help=f"Set Duckietown version. Use one of {allowed_branches}. Branches from "
             f"https://github.com/duckietown/duckietown-shell-commands of the form '[branch]-*' are also "
             f"supported.",
    )

    parsed, others = parser.parse_known_args(args)

    return CLIOptions(debug=parsed.debug, set_version=parsed.set_version, quiet=parsed.quiet), others



import sys

from setuptools import find_packages, setup


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


if sys.version_info < (3, 6):
    msg = 'duckietown-shell works with Python 3.6 and later.\nDetected %s.' % str(sys.version)
    sys.exit(msg)

distro = 'daffy'

shell_version = get_version(filename='lib/dt_shell/__init__.py')
install_requires = [
    'GitPython',
    'texttable',
    'base58',
    'ecdsa',
    'python-dateutil',
    'whichcraft',
    'termcolor',
    'PyYAML',
    'docker',
    'docker-compose',
    'six',
    'psutil',
    'future',
    'duckietown-docker-utils-{}>=6.0.9'.format(distro),
    'zeroconf',
    'requests',
    'dt-authentication-{}'.format(distro),
    'dt-data-api-{}>=0.1.6'.format(distro)
]

system_version = tuple(sys.version_info)[:3]
if system_version < (3, 7):
    install_requires.append('dataclasses')

setup(
    name='duckietown-shell',
    version=shell_version,
    download_url='http://github.com/duckietown/duckietown-shell/tarball/%s' % shell_version,
    package_dir={'': 'lib'},
    packages=find_packages('lib'),
    # we want the python 2 version to download it, and then exit with an error
    # python_requires='>=3.6',

    tests_require=[],
    install_requires=install_requires,
    # This avoids creating the egg file, which is a zip file, which makes our data
    # inaccessible by dir_from_package_name()
    zip_safe=False,

    # without this, the stuff is included but not installed
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'dts = dt_shell:cli_main',
        ]
    }
)
